# copy from module "game/prototype"
# time: 2024/03/20
from builtin.const import *


class UsrObject:
    """
    空对象, 用于添加自定义属性
    """
    pass


class Point(object):
    def __init__(self, x, y):
        """
        创建一个新的位置对象。

        Args:
            x (int): X 坐标
            y (int): Y 坐标

        Example:
            >>> pos = Point(10, 20)
            >>> print(pos)  # 输出: (10◈20)
        """
        self.x: int = x
        self.y: int = y

    @staticmethod
    def center(*point) -> "Point":
        """
        返回一组点的中心点。

        Args:
            *point (Point): 一组点

        Returns:
            Point: 中心点(使用round进行取整)
        """
        pass

    @staticmethod
    def fcenter(*point) -> "Point":
        """
        返回一组点的中心点。

        Args:
            *point (Point): 一组点

        Returns:
            Point: 中心点(不使用round进行取整)
        """
        pass

    def distance(self, target: "Point") -> int:
        """
        计算两个点之间的距离(切比雪夫距离)。

        Args:
            target (Point): 目标点

        Returns:
            int: 两个点之间的距离
        """
        pass

    def at(self, target: "Point") -> bool:
        """
        检查两个点是否相同。

        Args:
            target (Point): 目标点

        Returns:
            bool: 如果两个点相同，则返回 true，否则返回 false
        """
        pass

    def nat(self, target: "Point") -> bool:
        """
        检查两个点是否不同。

        Args:
            target (Point): 目标点

        Returns:
            bool: 如果两个点不同，则返回 true，否则返回 false
        """
        pass

    def offset(self, dx: int, dy: int) -> "Point":
        """
        返回一个新点，该点相对于当前点偏移了(dx, dy)。

        Args:
            dx (int): X轴偏移量
            dy (int): Y轴偏移量

        Returns:
            Point: 偏移后的新点
        """
        pass

    def closest(self, objs, filter_fn=None, notme=True):
        """
        找到距离当前点最近的对象。

        Args:
            objs (list): 对象列表
            filter_fn (callable, optional): 过滤函数。默认为 None。
            notme (bool, optional): 是否排除当前点本身。默认为 True。

        Returns:
            Point: 最近的对象，如果没有找到则返回 None
        """
        pass

    def quickest(self, objs, filter_fn=None, options=None, notme=True):
        """
        找到到达当前点最快的路径对象。

        Args:
            objs (list): 对象列表
            filter_fn (callable, optional): 过滤函数。默认为 None。
            options (dict, optional): 额外选项。默认为 None。
            notme (bool, optional): 是否排除当前点本身。默认为 True。

        Returns:
            Point: 最快的对象，如果没有找到则返回 None
        """
        pass

    def inrange(self, objs, range, filter_fn=None):
        """
        找到在指定范围内距离当前点的对象。

        Args:
            objs (list): 对象列表
            range (int): 范围
            filter_fn (callable, optional): 过滤函数。默认为 None。

        Returns:
            list: 在范围内的对象列表
        """
        pass


class Rect:
    def __init__(self, lt: Point | int, rb_or_w: Point | int, h: int | None = None, ltrb: int = None):
        """
        初始化矩形对象
        :param lt: 左上角点
        :param rb_or_w: 右下角点或宽度
        :param h: 高度
        :param ltrb: 如果这个参数被传入，那么会将所有参数视作int
        @example:
        Rect(Point(0,0), Point(10,10))
        or
        Rect(Point(0,0), 10, 10)
        or
        Rect(0,0,10,10)
        """
        self.left: int = 0
        self.top: int = 0
        self.right: int = 0
        self.bottom: int = 0

    def offset(self, dx: int, dy: int) -> "Rect":
        """
        平移矩形
        :param dx: x方向平移量
        :param dy: y方向平移量
        """
        pass

    @property
    def center(self) -> Point:
        """
        获取矩形中心点
        :return: Point对象，表示矩形中心点
        """
        return Point(0, 0)

    @property
    def width(self) -> int:
        """
        获取矩形宽度
        :return: 矩形宽度
        """
        return 0

    @property
    def height(self) -> int:
        """
        获取矩形高度
        :return: 矩形高度
        """
        return 0

    @property
    def area(self) -> int:
        """
        获取矩形面积
        :return: 矩形面积
        """
        return 0


class ClusterResult:
    x: int = 0
    y: int = 0
    grade: int = 0  # 如果聚类类型是Creep，那么会一并计算其得分; 否则其值为0
    objects: list[Point] = []
    pass


# CostMatrix 类
class CostMatrix:
    """
    自定义导航成本数据的容器。
    如果在CostMatrix中找到非0值，则将使用该值代替默认的地形成本。
    """

    def __init__(self):
        """
        创建一个新的CostMatrix，所有位置都包含0。
        """
        pass

    def get(self, x: int, y: int) -> int:
        """
        获取此CostMatrix中位置的成本。
        x: 游戏中的X位置
        y: 游戏中的Y位置
        返回指定位置的成本
        """
        pass

    def set(self, x: int, y: int, cost: int):
        """
        设置此CostMatrix中位置的成本。
        x: 游戏中的X位置
        y: 游戏中的Y位置
        cost: 此位置的成本。
        """
        pass

    def clone(self):
        """
        返回一个新的CostMatrix实例。
        """
        pass


class COST_MATRIX_LIBS:
    """
    基础的CostMatrix数据库。
    不了解的话请勿修改。

    用法:
        * 禁用某些位置，避免友军走到该位置(无法撤销)
        spawn = get.spawn(st.my)
        COST_MATRIX_LIBS.set(spawn.x - 1, spawn.y, 254) # 禁用spawn左侧一格
        COST_MATRIX_LIBS.set(spawn.x + 1, spawn.y, 254) # 禁用spawn右侧一格
    """
    default = CostMatrix()
    swamp = CostMatrix()

    @classmethod
    def set(cls, x: int, y: int, cost: int):
        """
        设置此CostMatrix中位置的成本。
        x: 游戏中的X位置
        y: 游戏中的Y位置
        cost: 此位置的成本。
        """
        pass

class MotionOptions:
    def __init__(self):
        """
        * 不可用户自行实例化
        costMatrix: 自定义导航成本数据
        plainCost: 在平原位置行走的成本，默认为2
        swampCost: 在沼泽位置行走的成本，默认为10
        flee: 是否寻找远离目标的路径，默认为False
        maxOps: 允许的最大寻路操作数，默认为50000
        maxCost: 返回路径的最大允许成本，默认为无穷大
        heuristicWeight: 应用于A*公式F = G + weight * H的启发式权重，默认为1.2
        """
        self.costMatrix: CostMatrix = None
        # 由于采用了自定义动态costMatrix，所以这两个参数已被禁用
        # self.plainCost = 2
        # self.swampCost = 10
        self.flee = False
        self.maxOps = 50000
        self.maxCost = float('inf')
        self.heuristicWeight = 1.2
        raise NotImplementedError(""
                                  "\n\n不可用户自行实例化, 如果想要编译，请运行build.py\nCannot be instantiated by users, if you want to compile, please run build.py")

    def clone(self):
        """
        返回一个新的MotionOptions实例。
        """
        pass


# 忽略沼泽地
DEFAULT_MOTION = MotionOptions()
SWAMP_MOTION = MotionOptions()


class PartsVector:
    """
    用于表示和操作 creep 的身体部件向量。
    """
    INDEXS = {
        "MOVE": 0,
        "CARRY": 1,
        "WORK": 2,
        "ATTACK": 3,
        "RANGED_ATTACK": 4,
        "HEAL": 5,
        "TOUGH": 6
    }

    def __init__(self, named_recipe):
        """
        初始化部件向量。

        Args:
            named_recipe (list[str] | NamedRecipe): creep 的身体部件配方
        """
        self.recipe = [""]
        self.vec7 = [0, 0, 0, 0, 0, 0, 0]

        self.nmCount = 0  # no move and carry
        self.bodyCount = 0  # body parts count

    def add(self, parts_vector_recipe):
        """
        将另一个部件向量添加到当前向量。

        Args:
            parts_vector_recipe (list[str] | PartsVector): 要添加的部件向量
        """
        pass

    def sub(self, parts_vector_recipe):
        """
        从当前向量中减去另一个部件向量。

        Args:
            parts_vector_recipe (list[str] | PartsVector): 要减去的部件向量
        """
        pass

    @property
    def others(self) -> int:
        """
        获取非移动和非搬运部件的数量。

        Returns:
            int: 非移动和非搬运部件的数量
        """
        return 0

    @property
    def moves(self) -> int:
        """
        获取移动部件的数量。

        Returns:
            int: 移动部件的数量
        """
        return 0

    @property
    def carries(self) -> int:
        """
        获取搬运部件的数量。

        Returns:
            int: 搬运部件的数量
        """
        return 0

    @property
    def works(self) -> int:
        """
        获取工作部件的数量。

        Returns:
            int: 工作部件的数量
        """
        return 0

    @property
    def melees(self) -> int:
        """
        获取近战攻击部件的数量。

        Returns:
            int: 近战攻击部件的数量
        """
        return 0

    @property
    def ranges(self) -> int:
        """
        获取远程攻击部件的数量。

        Returns:
            int: 远程攻击部件的数量
        """
        return 0

    @property
    def heals(self) -> int:
        """
        获取治疗部件的数量。

        Returns:
            int: 治疗部件的数量
        """
        return 0

    @property
    def toughs(self) -> int:
        """
        获取坚韧部件的数量。

        Returns:
            int: 坚韧部件的数量
        """
        return 0

    @staticmethod
    def similarity(a: "PartsVector", b: "PartsVector") -> float:
        """
        计算两个部件向量的相似度。

        Args:
            a (PartsVector): 第一个部件向量
            b (PartsVector): 第二个部件向量

        Returns:
            float: 两个部件向量的相似度
        """
        pass

    @staticmethod
    def selectVector(a: "PartsVector", vecs: list["PartsVector"]) -> "PartsVector":
        """
        从一组部件向量中选择与给定向量最相似的向量。

        Args:
            a (PartsVector): 给定向量
            vecs (list[PartsVector]): 部件向量列表

        Returns:
            PartsVector: 与给定向量最相似的向量
        """
        pass

    @staticmethod
    def parts(creep: "Creep", broken=True) -> list[str]:
        """
        获取 creep 的身体部件。

        Args:
            creep (Creep): creep 对象
            broken (bool, optional): 是否包括损坏的部件，默认为 True

        Returns:
            list[str]: creep 的身体部件列表
        """
        pass

    SCORE_TABLE = {
        "ATTACK": 10,
        "RANGED_ATTACK": 15,
        "HEAL": 15,
        "TOUGH": 1,
        "MOVE": 2,
        "CARRY": 2,
        "WORK": 5,
    }

    @staticmethod
    def partsGrade(creep_recipe: list[str]) -> int:
        """
        计算 creep 配方的评分。

        Args:
            creep_recipe (list[str] | Creep): creep 的身体部件配方或 creep 对象

        Returns:
            int: creep 配方的评分
        """
        pass

    @staticmethod
    def partsOptimise(parts: list[str]) -> list[str]:
        """
        优化 creep 的身体部件配方。

        Args:
            parts (list[str]): creep 的身体部件配方

        Returns:
            list[str]: 优化后的 creep 身体部件配方
        """
        pass

    PARTS_COST = {
        "MOVE": 50,
        "WORK": 100,
        "CARRY": 50,
        "ATTACK": 80,
        "RANGED_ATTACK": 150,
        "HEAL": 250,
        "TOUGH": 10,
    }

    @staticmethod
    def partsCost(creep_recipe: list[str] | "Creep") -> int:
        """
        计算 creep 配方的成本。

        Args:
            creep_recipe (list[str] | Creep): creep 的身体部件配方或 creep 对象

        Returns:
            int: creep 配方的成本
        """
        pass


class NamedRecipe:
    """
    用于表示和管理命名的 creep 配方。
    """

    def __init__(self, name: str, creep_recipe: list[str] | "Creep"):
        """
        初始化命名的 creep 配方。

        Args:
            name (str): 配方的名称
            creep_recipe (list[str] | GameCreep): creep 的身体部件配方或 GameCreep 对象
        """

        self.name = name
        self.recipe = [""]
        self.grade: int = 0
        self.cost: int = 0
        self.effect: float = 0.5


class CreepInfo:
    """
    用于表示 creep 的信息。
    """
    ATTACK_POWER = 30
    HEAL_POWER = 12
    RANGED_ATTACK_POWER = 10

    def __init__(self, creep):
        """
        初始化 creep 信息。

        Args:
            creep: creep 对象
        """
        self.creep = creep
        self.name = ""
        self.namedRecipe = NamedRecipe(self.name, creep)
        self.partsVector = PartsVector(self.namedRecipe)
        self.grade = self.namedRecipe.grade
        self.cost = self.namedRecipe.cost
        self.effect = self.namedRecipe.effect

        # 特性描述
        self.melee: bool = True
        self.ranged: bool = True
        self.heal: bool = True
        self.work: bool = True
        self.storable: bool = True

    @property
    def dynamicPartsVector(self) -> PartsVector:
        """
        获取动态部件向量。
        """
        return PartsVector(self.namedRecipe)

    @property
    def attackPower(self) -> int:
        """
        获取 creep 的攻击力。

        比如3个ATTACK部件和1个RANGE_ATTACK部件，那么攻击力为3 * 30 + 1 * 10 = 100
        """
        return 0

    @property
    def dynamicAttackPower(self) -> int:
        """
        获取 creep 的动态攻击力。

        参考目前creep剩余的部件，计算攻击力。
        """
        return 0

    @property
    def meleePower(self) -> int:
        """
        获取 creep 的近战力。

        比如3个ATTACK部件和1个RANGE_ATTACK部件，那么近战力为3 * 30 = 90
        """
        return 0

    @property
    def dynamicMeleePower(self) -> int:
        """
        获取 creep 的动态近战力。

        参考目前creep剩余的部件，计算近战力。
        """
        return 0

    @property
    def rangedPower(self) -> int:
        """
        获取 creep 的远程攻击力。

        比如3个ATTACK部件和1个RANGE_ATTACK部件，那么远程攻击力为1 * 10 = 10
        """
        return 0

    @property
    def dynamicRangedPower(self) -> int:
        """
        获取 creep 的动态远程攻击力。

        参考目前creep剩余的部件，计算远程攻击力。
        """
        return 0

    @property
    def healPower(self) -> int:
        """
        获取 creep 的治疗力。

        比如3个HEAL部件，那么治疗力为3 * 12 = 36
        """
        return 0

    @property
    def dynamicHealPower(self) -> int:
        """
        获取 creep 的动态治疗力。

        参考目前creep剩余的部件，计算治疗力。
        """
        return 0

    @property
    def motionAbility(self) -> float:
        """
        获取 creep 的移动率。

        考虑两个因素，MOVE组件的数量(记作moves)，和除开MOVE和CARRY后 剩余组件的数量(记作others)。
        移动率 = moves * 2 / (others * 10)
        """
        return 0.0

    @property
    def dynamicMotionAbility(self) -> float:
        """
        获取 creep 的动态移动率。

        参考目前creep剩余的部件，计算移动率。
        """
        return 0.0

    @property
    def armorRatio(self) -> float:
        """
        获取 creep 的装甲率。

        装甲率 = TOUGH组件的数量 / 所有组件的数量
        """
        return 0.0

    @property
    def dynamicArmorRatio(self) -> float:
        """
        获取 creep 的动态装甲率。

        参考目前creep剩余的部件，计算装甲率。
        """
        return 0.0


class CreepMotion:
    """
    对己方队友，记录最近一次移动的操作
    对敌方敌人，追踪并预测其位置
    """
    SEPARATE_LENGTH = 3  # 每次长度的便取一个点，作为阶段性移动的目标点

    def __init__(self, creep):
        """
        初始化 CreepMotion 对象。

        Args:
            creep: creep 对象
        """
        self.creep = creep
        self.path = []  # 记录已有的一组路径，让 creep 沿着这条路径移动
        self.path_goal = {"x": creep.x, "y": creep.y}  # 记录路径的终点
        self.path_index = 0  # 记录路径的索引
        self.lastPoint = {"x": creep.x, "y": creep.y}  # 记录敌方 creep 上一帧的位置

    def seekNext(self, options: MotionOptions = None) -> Point:
        """
        对友军，寻找下一个目标点
        对敌方，寻找下一个预测点

        Args:
            options (dict, optional): 只对友军有效，作为没有路径时的寻路选项
        """
        pass

    def seekNextStagePoint(self, new_path_goal: Point, options: MotionOptions = None) -> Point:
        """
        只对友军有效，寻找下一个阶段性目标点

        Args:
            new_path_goal (dict): 新的目标点
            options (dict, optional): 寻路选项
        """
        pass

    def clear(self):
        """
        清空所有信息
        """
        pass


# GameObject 类
class GameObject(Point):
    """
    游戏对象的基本原型。
    """

    # 类属性，用于存储前缀表
    PREFIX_TABLE = {
        'Creep': "c",
        'StructureSpawn': "sp",
        'StructureExtension': "ext",
        'StructureRoad': "road",
        'StructureWall': "wall",
        'StructureRampart': "ramp",
        'StructureContainer': "box",
        'StructureTower': "twr",
        'Structure': "st",
        'ScoreCollector': "sc",
        'AreaEffect': "ae",
        'Source': "sr",
        'ConstructionSite': "site",
        'Flag': "flag",
        'Resource': "rs",
        'BodyPart': "bp",
        'GameObject': "go",
        'Object': "o",
    }

    def __init__(self, gobject):
        """
        游戏中的对象
        * 这个类不能被你实例化
        """
        super().__init__(0, 0)
        self.id = "1"
        self.name = "go1"
        self.my = True
        self.ticksToDecay = 0
        raise NotImplementedError

    @property
    def exists(self) -> bool:
        """
        检查对象是否存在。
        """
        return True


class DynamicBody:
    """
    动态身体部件类。
    """

    def __init__(self):
        self.type = MOVE
        self.hits = 0


class Creep(GameObject):
    """
    代表游戏中的单位（Creep）。
    """

    def __init__(self, creep):
        """
        初始化 Creep 对象。

        Args:
            creep: creep 对象
        """
        super().__init__(creep)
        self.motion:CreepMotion = CreepMotion(creep)
        self.info:CreepInfo = CreepInfo(creep)
        self.grade:int = self.info.grade
        self.recipe:list[str] = self.info.namedRecipe.recipe
        self.partsVector:PartsVector = self.info.partsVector

    # 记录最近一个tick的操作内容
    class actions:
        move = None  # Point
        attack = None  # Creep | Structure
        heal = None  # Creep
        harvest = None  # Source
        build = None  # Site
        fetch = None  # Resource | Structure | Creep
        deposit = None  # Creep | Structure | Point

    @property
    def birthed(self) -> bool:
        """
        检查 creep 是否已经出生。
        """
        return True

    @property
    def fatigue(self) -> int:
        """
        获取 creep 的疲劳值。
        """
        return 0

    @property
    def energy(self) -> int:
        """
        获取 creep 的能量值。
        """
        return 0

    @property
    def energyMax(self) -> int:
        """
        获取 creep 的最大能量值。
        """
        return 0

    @property
    def hits(self) -> int:
        """
        获取 creep 的当前生命值。
        """
        return 0

    @property
    def hitsMax(self) -> int:
        """
        获取 creep 的最大生命值。
        """
        return 0

    @property
    def hp(self) -> int:
        """
        获取 creep 的当前生命值。
        """
        return 0

    @property
    def hpMax(self) -> int:
        """
        获取 creep 的最大生命值。
        """
        return 0

    @property
    def store(self) -> "Store":
        """
        获取 creep 的存储信息。
        """
        return Store("")

    @property
    def body(self) -> list[DynamicBody]:
        """
        获取 creep 的身体部件信息。
        """
        return []

    def listenFor(self, event_name: str, callback: callable, sources: list[str | GameObject] | None = None) -> None:
        """
        监听事件。

        Args:
            event_name (str): 事件名称
            callback (callable): 回调函数
            sources (list[str|GameObject], optional): 事件源限制，只会监听到这些对象的事件。默认为 None
        """
        pass

    def pushEvent(self, event_name: str, data: dict | object) -> None:
        """
        推送事件。

        Args:
            event_name (str): 事件名称
            data (dict|object): 事件数据
        """
        pass

    def cancelListen(self, event_name: str) -> None:
        """
        取消监听事件。

        Args:
            event_name (str): 事件名称
        """
        pass

    def schedule(self, delay: int, callback: callable) -> None:
        """
        安排任务。

        Args:
            delay (int): 延迟时间
            callback (callable): 回调函数
        """
        pass

    def move(self, to: Point | int, options: MotionOptions = None) -> int:
        """
        移动 creep。

        Args:
            to (Point|int): 目标位置或者方向常量
            options (MotionOptions, optional): 移动选项

        Returns:
            int: 错误码
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_ARGS	-10	The provided direction is incorrect.
            ERR_TIRED	        -11	The fatigue indicator of the creep is non-zero.
            ERR_NO_BODYPART	    -12	There are no MOVE body parts in this creep’s body.

        """
        pass

    def escape(self, target: Point, options: MotionOptions = None) -> int:
        """
        逃离目标。

        Args:
            target (Point): 目标对象
            options (MotionOptions, optional): 移动选项

        Returns:
            int: 错误码
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_ARGS	-10	The provided direction is incorrect.
            ERR_TIRED	        -11	The fatigue indicator of the creep is non-zero.
            ERR_NO_BODYPART	    -12	There are no MOVE body parts in this creep’s body.
        """
        pass

    def attack(self, target: "Creep" | "Structure", move: MotionOptions | bool = True) -> int:
        """
        攻击目标。

        Args:
            target (Creep, Structure): 目标对象
            move (bool, MotionOptions): 是否移动。默认为 True

        Returns:
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_TARGET	-7	The target is not a valid attackable object.
            ERR_NOT_IN_RANGE	-9	The target is too far away.
            ERR_NO_BODYPART	    -12	There are no (RANGED_)ATTACK body parts in this creep’s body.
        """
        pass

    def heal(self, target: "Creep", move: MotionOptions | bool = True) -> int:
        """
        治疗目标。

        Args:
            target (Creep): 目标对象
            move (bool, MotionOptions): 是否移动。默认为 True

        Returns:
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_TARGET	-7	The target is not a valid creep object.
            ERR_NOT_IN_RANGE	-9	The target is too far away.
            ERR_NO_BODYPART	    -12	There are no HEAL body parts in this creep’s body.

        """
        pass

    def fetch(self, target: "Creep" | "Resource" | "Structure", resource_type: str = None, amount: int = None, move: bool | MotionOptions = True) -> int:
        """
        从目标获取资源。

        Args:
            target (Creep, Resource, Structure): 目标对象
            resource_type (str, optional): 资源类型。默认为 None, 被解释为RESOURCE_ENERGY
            amount (int, optional): 资源数量。默认为 None, 表示没有数量限制
            move (bool, MotionOptions): 是否移动。默认为 True

        Returns:
            OK	                        0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	            -1	You are not the owner of this creep.
            ERR_NOT_ENOUGH_RESOURCES	-6	The creep does not have the given amount of resources.
            ERR_INVALID_TARGET	        -7	The target is not a valid creep object.
            ERR_FULL	                -8	The creep cannot receive any more resource.
            ERR_NOT_IN_RANGE	        -9	The target is too far away.
            ERR_INVALID_ARGS	        -10	The resourceType is not a valid RESOURCE_* constant.
        """
        pass

    def deposit(self, target: "Creep" | "Structure" | Point | None, resource_type: str = None, amount: int = None, move: bool | MotionOptions = True) -> int:
        """
        将资源存储到目标。

        Args:
            target (Creep, Structure, Point, None): 目标对象, None表示放置在原地, Point表示放置在指定位置
            resource_type (str, optional): 资源类型。默认为 None, 被解释为RESOURCE_ENERGY
            amount (int, optional): 资源数量。默认为 None, 表示没有数量限制
            move (bool, MotionOptions): 是否移动。默认为 True

        Returns:
            OK	                        0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	            -1	You are not the owner of this creep.
            ERR_NOT_ENOUGH_RESOURCES	-6	The creep does not have the given amount of resources.
            ERR_INVALID_TARGET	        -7	The target is not a valid creep object.
            ERR_FULL	                -8	The target cannot receive any more resource.
            ERR_NOT_IN_RANGE	        -9	The target is too far away.
            ERR_INVALID_ARGS	        -10	The resourceType is not a valid RESOURCE_* constant.
        """
        pass

    def build(self, site: "ConstructionSite", move: bool | MotionOptions = True) -> int:
        """
        建造结构。

        Args:
            site(ConstructionSite): 建造位置
            move (bool, MotionOptions): 是否移动。默认为 True

        Returns:
            OK	                        0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	            -1	You are not the owner of this creep.
            ERR_NOT_ENOUGH_RESOURCES	-6	The creep does not have any carried energy.
            ERR_INVALID_TARGET	        -7	The target is not a valid construction site object or the structure cannot be built here (probably because of an obstacle at the same square).
            ERR_NOT_IN_RANGE	        -9	The target is too far away.
            ERR_NO_BODYPART	            -12	There are no WORK body parts in this creep’s body.
        """
        pass

    def harvest(self, target: "Source", move: MotionOptions | bool = True) -> int:
        """
        采集资源。

        Args:
            target (Source): 采集目标(资源点)
            move (bool, MotionOptions): 是否移动。默认为 True

        Returns:
            OK	                        0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	            -1	You are not the owner of this creep.
            ERR_NOT_ENOUGH_RESOURCES	-6	The target does not contain any harvestable resource.
            ERR_INVALID_TARGET	        -7	The target is not a valid source object.
            ERR_NOT_IN_RANGE	        -9	The target is too far away.
            ERR_NO_BODYPART	            -12	There are no WORK body parts in this creep’s body.
        """
        pass

    def intermit(self, target: "Creep" | "Structure" | Point, resource_type: str = None, options: MotionOptions = None):
        """
        以“断断续续”的方式搬运资源。

        Args:
            target (Creep, Structure, Point): 目标对象, Point表示放置在指定位置
            resource_type (str, optional): 资源类型。默认为 None, 被解释为RESOURCE_ENERGY
            options (dict, MotionOptions): 移动选项。默认为 None

        Returns:
            OK	                        0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	            -1	You are not the owner of this creep.
            ERR_NOT_ENOUGH_RESOURCES	-6	The creep does not have the given amount of resources.
            ERR_INVALID_TARGET	        -7	The target is not a valid creep object.
            ERR_FULL	                -8	The target cannot receive any more resource.
            ERR_NOT_IN_RANGE	        -9	The target is too far away.
            ERR_INVALID_ARGS	        -10	The resourceType is not a valid RESOURCE_* constant.
        """
        pass

    def follow(self, target: GameObject, distance: int | tuple, options: MotionOptions = None) -> int:
        """
        跟随目标。

        Args:
            target (GameObject): 目标对象
            distance (int or tuple): 与目标保持的距离
                    * target是友方时，才支持tuple参数，表示(`stop_dist`, `ignore_dist`)
                    * 超出stop_dist时，leader向follower移动，直到距离为stop_dist
                    * 超出ignore_dist时，leader忽略follower，follower自行移动
            options (dict, MotionOptions): 移动选项。默认为 None

        Returns:
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_ARGS	-10	The provided direction is incorrect.
            ERR_TIRED	        -11	The fatigue indicator of the creep is non-zero.
            ERR_NO_BODYPART	    -12	There are no MOVE body parts in this creep’s body.
        """
        pass

    def carry(self, src: "Creep" | "Resource" | "Structure", dst: "Creep" | "Structure" | Point, resource_type: str = None, options: MotionOptions = None, intermit: bool = False) -> int:
        """
        搬运资源。

        Args:
            src (Creep, Resource, Structure): 源对象
            dst (Creep, Structure, Point): 目标对象, Point表示放置在指定位置
            resource_type (str, optional): 资源类型。默认为 None
            options (dict, MotionOptions): 移动选项。默认为 None
            intermit (bool, optional): 是否采用“断断续续”的搬运方式。默认为 False

        Returns:
            DONE                        1	The operation has been completed successfully.
            OK	                        0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	            -1	You are not the owner of this creep.
            ERR_NOT_ENOUGH_RESOURCES	-6	The creep does not have the given amount of resources.
            ERR_INVALID_TARGET	        -7	The target is not a valid creep object.
            ERR_FULL	                -8	The creep cannot receive any more resource.
            ERR_NOT_IN_RANGE	        -9	The target is too far away.
            ERR_INVALID_ARGS	        -10	The resourceType is not a valid RESOURCE_* constant.
        """
        pass

    def patrol(self, points: list[Point], options: MotionOptions = None) -> int:
        """
        巡逻。

        Args:
            points (list): 巡逻点列表
            options (MotionOptions, optional): 移动选项。默认为 None

        Returns:
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_ARGS	-10	The provided direction is incorrect.
            ERR_TIRED	        -11	The fatigue indicator of the creep is non-zero.
            ERR_NO_BODYPART	    -12	There are no MOVE body parts in this creep’s body.
        """
        pass

    def pull(self, target: "Creep", options:MotionOptions=None) -> int:
        """
        将目标拉向你，如果距离大于1，则会先移动到目标旁边。
        * 你可以决定你要去往何处

        Args:
            target (Creep): 目标友方对象
            options (MotionOptions, optional): 移动选项。默认为 None

        Returns:
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_TARGET	-7	The target is not a valid creep object.
            ERR_NOT_IN_RANGE	-9	The target is too far away.
        """
        pass

    def push(self, target: "Creep", options:MotionOptions=None) -> int:
        """
        推动目标, 如果距离大于1，则会先移动到目标旁边。
        * 目标必须具有移动能力，你不能决定你要去往何处

        Args:
            target (Creep): 目标友方对象
            options (MotionOptions, optional): 移动选项。默认为 None

        Returns:
            OK	                0	The operation has been scheduled successfully.
            ERR_NOT_OWNER	    -1	You are not the owner of this creep.
            ERR_INVALID_TARGET	-7	The target is not a valid creep object.
            ERR_NOT_IN_RANGE	-9	The target is too far away.
        """
        pass

    def autoAttack(self, view_range: int = 50, wall:bool=True) -> tuple[GameObject | None, GameObject | None]:
        """
        自动攻击
        * 会自动搜索攻击范围内的敌人，并返回锁定的目标和攻击目标
        * 如果没有敌人，则返回 None

        Args:
            view_range (int, optional): 搜索范围，默认为 50
            wall (bool, optional): 是否攻击墙壁，默认为 True

        Returns:
            tuple[GameObject|None, GameObject|None]: 锁定的目标和攻击目标
        """
        pass

    def autoHeal(self, view_range: int = 50) -> tuple[GameObject | None, GameObject | None]:
        """
        自动治疗
        * 会自动搜索治疗范围内的友方单位，并返回锁定的目标和治疗目标
        * 如果没有友方单位，则返回 None

        Args:
            view_range (int, optional): 搜索范围，默认为 50

        Returns:
            tuple[GameObject|None, GameObject|None]: 锁定的目标和治疗目标
        """
        pass


class Structure(GameObject):
    """
    代表游戏中的结构体对象。
    """

    def __init__(self, gobject):
        """
        初始化结构体对象。
        * 这个类不能被你实例化

        Args:
            gobject: 包含结构体属性的对象
        """
        super().__init__(gobject)
        self.my = gobject.my
        self.hitsMax = gobject.hitsMax

    @property
    def hits(self) -> int:
        """
        获取当前结构体的耐久值。

        Returns:
            int: 当前耐久值
        """
        return 0

    @property
    def hp(self) -> int:
        """
        获取 当前结构体的生命值。
        """
        return 0

    @property
    def hpMax(self) -> int:
        """
        获取 结构体的最大生命值。
        """
        return 0


class Resource(GameObject):
    """
    代表游戏中的资源对象。
    """

    def __init__(self, gobject):
        """
        初始化资源对象。
        * 这个类不能被你实例化

        Args:
            gobject: 包含资源属性的对象
        """
        super().__init__(gobject)
        self.resourceType = gobject.resourceType

    @property
    def amount(self) -> int:
        """
        获取资源的数量。

        Returns:
            int: 资源的数量
        """
        return 0

    @property
    def energy(self) -> int:
        """
        获取 资源的能量值。
        """
        return 0


class Source(GameObject):
    """
    代表游戏中的资源源对象（例如能量源）。
    """

    def __init__(self, gobject):
        """
        初始化资源源对象。
        * 这个类不能被你实例化

        Args:
            gobject: 包含资源源属性的对象
        """
        super().__init__(gobject)
        self.energyCapacity = gobject.energyCapacity

    @property
    def energy(self) -> int:
        """
        获取当前资源源的能量值。

        Returns:
            int: 当前能量值
        """
        return 0

    @property
    def energyMax(self) -> int:
        """
        获取 资源源的最大能量值。
        """
        return 0


class StructureContainer(Structure):
    """
    代表游戏中的容器结构（StructureContainer）。
    """

    def __init__(self, gobject):
        """
        初始化容器结构对象。
        * 这个类不能被你实例化

        Args:
            gobject: 包含容器结构属性的对象
        """
        super().__init__(gobject)
        self.cost = 200
        self.capacity = 2000

    @property
    def store(self):
        """
        获取容器中的存储内容。

        Returns:
            dict: 容器中的存储内容（例如资源类型和数量）
        """
        return Store("")

    @property
    def energy(self) -> int:
        """
        获取 容器中的能量值。
        """
        return 0

    @property
    def energyMax(self) -> int:
        """
        获取 容器的最大能量值。
        """
        return 0


class StructureExtension(StructureContainer):
    """
    代表游戏中的扩展结构（StructureExtension）。
    """

    def __init__(self, gobject):
        """
        初始化扩展结构对象。
        * 这个类不能被你实例化

        Args:
            gobject: 包含扩展结构属性的对象
        """
        super().__init__(gobject)
        self.cost = 200
        self.capacity = 100


class ScoreCollector(GameObject):
    """
    代表游戏中的得分收集器对象（ScoreCollector）。
    """

    def __init__(self, gobject):
        """
        初始化得分收集器对象。

        Args:
            gobject: 包含得分收集器属性的对象
        """
        super().__init__(gobject)
        self.capacity = gobject.scoreTotal
        self.scoreTotal = gobject.scoreTotal
        self.resourceType = RESOURCE_SCORE  # 假设 RESOURCE_SCORE 是一个常量

    @property
    def my(self) -> bool:
        """
        获取得分收集器是否属于当前玩家。

        Returns:
            bool: 是否属于当前玩家
        """
        return True

    @property
    def score(self) -> int:
        """
        获取当前得分。

        Returns:
            int: 当前得分
        """
        return 0


class AreaEffect(GameObject):
    """
    代表游戏中的区域效果对象（AreaEffect）。
    """

    def __init__(self, gobject):
        """
        初始化区域效果对象。

        Args:
            gobject: 包含区域效果属性的对象
        """
        super().__init__(gobject)
        self.effect = gobject.effect


class StructureWall(Structure):
    """
    代表游戏中的墙壁结构（StructureWall）。
    """

    def __init__(self, gobject):
        """
        初始化墙壁结构对象。

        Args:
            gobject: 包含墙壁结构属性的对象
        """
        super().__init__(gobject)
        self.cost = 100


class StructureRampart(StructureWall):
    """
    代表游戏中的城墙（StructureRampart）。
    """

    def __init__(self, gobject):
        """
        初始化城墙对象。

        Args:
            gobject: 包含城墙属性的对象
        """
        super().__init__(gobject)
        self.cost = 200


class StructureRoad(StructureWall):
    """
    代表游戏中的道路结构（StructureRoad）。
    """

    def __init__(self, gobject):
        """
        初始化道路结构对象。

        Args:
            gobject: 包含道路结构属性的对象
        """
        super().__init__(gobject)
        self.cost = 50


class StructureTower(StructureContainer):
    """
    代表游戏中的塔结构（StructureTower）。
    """

    def __init__(self, gobject):
        """
        初始化塔结构对象。

        Args:
            gobject: 包含塔结构属性的对象
        """
        super().__init__(gobject)
        self.cost = 1250
        self.capacity = 50

    @property
    def cooldown(self) -> int:
        """
        获取塔的冷却时间。

        Returns:
            int: 当前冷却时间
        """
        return 0

    def measure(self, target) -> int:
        """
        计算对目标的伤害或是治疗量。

        Args:
            target: 目标对象

        Returns:
            int: 伤害或治疗量
        """
        pass

    def attack(self, target) -> int:
        """
        对目标发动攻击。

        Args:
            target: 目标对象

        Returns:
            int: 攻击结果
        """
        pass

    def heal(self, target) -> int:
        """
        对目标进行治疗。

        Args:
            target: 目标对象

        Returns:
            int: 治疗结果
        """
        pass


class StructureSpawn(StructureContainer):
    """
    代表游戏中的出生点结构（StructureSpawn）。
    """

    def __init__(self, gobject):
        """
        初始化出生点结构对象。

        Args:
            gobject: 包含出生点结构属性的对象
        """
        super().__init__(gobject)
        self.cost = 3000
        self.capacity = 1000

    @property
    def spawning(self) -> "Spawning":
        """
        获取当前是否正在孵化。

        Returns:
            spawning: 当前孵化对象，如果不存在则返回 None
        """
        return Spawning(0, 0, None)

    @property
    def needTime(self) -> int:
        """
        获取孵化所需的总时间。

        Returns:
            int: 孵化所需的总时间
        """
        return 0

    @property
    def remainingTime(self) -> int:
        """
        获取孵化剩余时间。

        Returns:
            int: 孵化剩余时间
        """
        return 0

    @property
    def creep(self) -> "Creep":
        """
        获取当前孵化中的 creep。

        Returns:
            Creep: 当前孵化中的 creep，如果不存在则返回 None
        """
        pass

    def cancel(self):
        """
        取消当前孵化。

        Returns:
            int: 错误代码，如果成功则返回 OK
        """
        pass

    def create(self, recipe, creep_name=None, optimise=True, direction=None):
        """
        使用指定的配方创建一个 creep。

        Args:
            recipe (list[str]): 用于创建 creep 的身体部位数组
            creep_name (str, optional): 可选的 creep 名称，默认为 None
            optimise (bool, optional): 是否优化配方以提高功能完存率，默认为 True
            direction (int, optional): 可选的方向参数，用于指定 creep 诞生时的方向，默认为 None

        Returns:
            Creep|number: 返回新创建的 creep 或剩余的孵化时间(>0)，或错误代码(<0)
        """
        pass


class ConstructionSite(GameObject):
    """
    代表游戏中的建筑工地（ConstructionSite）。
    """

    def __init__(self, gobject, building_type):
        """
        初始化建筑工地对象。

        Args:
            gobject: 包含建筑工地属性的对象
            building_type: 建筑类型
        """
        super().__init__(gobject)
        self.my: bool = gobject.my
        self.type: type = building_type
        self.progressTotal: int = gobject.progressTotal

    @property
    def progress(self) -> int:
        """
        获取建筑进度。
        """
        return 0

    @property
    def structure(self) -> Structure:
        """
        获取建筑结构。
        """
        return Structure(None)

    @property
    def percentage(self) -> float:
        """
        获取建筑进度的百分比。
        """
        return 0

    @staticmethod
    def plan(x: int, y: int, building_type: type, single: bool = False) -> int | "ConstructionSite":
        """
        在指定位置规划建筑工地。

        Args:
            x (int): 建筑工地的 x 坐标
            y (int): 建筑工地的 y 坐标
            building_type (type): 建筑类型
            single (bool, optional): 是否只允许单个建筑工地，默认为 False(允许重叠)

        Returns:
            int: 错误代码，如果成功则返回ConstructionSite对象
        """
        pass

    def remove(self):
        """
        移除建筑工地。
        """
        pass


class Flag(GameObject):
    """
    代表游戏中的旗帜（Flag）。
    """

    def __init__(self, gobject):
        """
        初始化旗帜对象。

        Args:
            gobject: 包含旗帜属性的对象
        """
        super().__init__(gobject)
        self.my: bool = gobject.my


class BodyPart(GameObject):
    """
    代表游戏中的身体部件（BodyPart）。
    """

    def __init__(self, gobject):
        """
        初始化身体部件对象。

        Args:
            gobject: 包含身体部件属性的对象
        """
        super().__init__(gobject)
        self.type: str = gobject.type


# Store 类型
class Store:
    def __init__(self, resources: str):
        self.resources = resources

    def getCapacity(self, resource_type: str = None) -> int | None:
        """
        返回此存储对指定资源的容量
        """
        pass

    def getUsedCapacity(self, resource_type: str = None) -> int | None:
        """
        返回存储使用的容量
        """
        pass

    def getFreeCapacity(self, resource_type: str = None) -> int | None:
        """
        返回存储的剩余容量
        """
        pass


# Spawning 类
class Spawning:
    def __init__(self):
        """
        needTime: 完成孵化所需的总时间
        remainingTime: 剩余时间
        creep: 正在孵化的 creep
        """
        self.needTime: int = 0
        self.remainingTime: int = 0
        self.creep: Creep = None

    def cancel(self) -> int:
        """
        立即取消孵化

        Returns:
            int: 错误代码，如果成功则返回 OK

        """
        pass


class GlobalKnowledge:
    """
    全局知识每tick更新器，用于跟踪和分类游戏中的所有对象，包括creep、结构和战斗单位。

    该类通过将对象分为敌对和友方，为游戏中的各种实体提供了一个集中的数据结构。
    它使游戏逻辑能够高效地访问和处理这些实体。

    Attributes:
        now (int): 当前游戏时间戳。

    内部类：
        battles: 每tick更新所有战斗单位，分为敌对和友方。
        civilian: 每tick更新所有民用单位，分为敌对和友方。
        clusters: 每tick更新所有creep集群，分为敌对和友方。

    """
    now = 0

    class battles:
        """
        每tick更新游戏中的所有战斗单位，分为敌对和友方。

        Attributes:
            enemies (list[Creep]): 敌对战斗单位列表。
            friends (list[Creep]): 友方战斗单位列表。
        """
        enemies: list[Creep] = []
        friends: list[Creep] = []

    class civilian:
        """
        每tick更新游戏中的所有民用单位，分为敌对和友方。

        Attributes:
            enemies (list[Creep]): 敌对民用单位列表。
            friends (list[Creep]): 友方民用单位列表。
        """
        enemies: list[Creep] = []
        friends: list[Creep] = []

    class clusters:
        """
        每tick更新游戏中的所有creep集群，分为敌对和友方。

        Attributes:
            enemies (list[ClusterResult]): 敌对creep集群列表。
            friends (list[ClusterResult]): 友方creep集群列表。
        """
        enemies: list[ClusterResult] = []
        friends: list[ClusterResult] = []


P = Point
know = GlobalKnowledge()
