from builtin.stdio import *


class View:
    """
    Visual的包装类

    Args:
        layer: int  视图层级
        persistent: bool  是否持久化视图

    Examples:
        ```python
        view = ut.View(1, True)
        view.clear().text('Hello', st.point(0, 0), 0.5, '#FF0000')  # > ignore
        ```
    """

    def __init__(self, layer: int, persistent: bool = True):
        pass

    def clear(self) -> object:
        """
        清空当前视图

        * 如果编译器提出警告，可以使用 `# > ignore` 来忽略

        Returns:
            View
        """
        pass

    def text(self, text: str, pos: st.point, font: str | int | float, color: str, options: UsrObject = None) -> object:
        """
        绘制文字

        Args:
            text: str  文字内容
            pos: st.point  文字位置
            font: str | int | float  字体大小 可以是 0.5 '12px' 这样的 (优先级高于options.font)
            color: str  颜色  比如 '#FF0000' (优先级高于options.color)
            options: 详见ScreepsArena文档

        Returns:
            View
        """
        pass

    def circle(self, pos: st.point, radius: float, fill: str = None, options: UsrObject = None) -> object:
        """
        绘制圆形

        Args:
            pos: st.point  圆心坐标
            radius: float  半径 (单位为方块) (优先级高于options.radius)
            fill: str  填充颜色  比如 '#FF0000' (优先级高于options.fill)
            options: 详见ScreepsArena文档

        Returns:
            View
        """
        pass

    def line(self, pos1: st.point, pos2: st.point, width: float = 0.1, color: str = None, options: UsrObject = None) -> object:
        """
        绘制线段

        Args:
            pos1: st.point  起点坐标
            pos2: st.point  终点坐标
            width: float  线宽 (单位为方块) (优先级高于options.width)
            color: str  颜色 比如 '#FF0000' (优先级高于options.color)
            options: 详见ScreepsArena文档

        Returns:
            View
        """
        pass

    def rect(self, pos: st.point, width: float, height: float, fill: str = None, options: UsrObject = None) -> object:
        """
        绘制矩形

        Args:
            pos: st.point  左上角坐标
            width: float  宽度
            height: float  高度
            fill: str  填充颜色 比如 '#FF0000' (优先级高于options.fill)
            options: 详见ScreepsArena文档

        Returns:
            View
        """
        pass

    def poly(self, points: list[st.point], fill: str = None, options: UsrObject = None) -> object:
        """
        绘制多边形

        Args:
            points: list[st.point]  多边形的顶点坐标
            fill: str  填充颜色 比如 '#FF0000' (优先级高于options.fill)
            options: 详见ScreepsArena文档

        Returns:
            View
        """
        pass

    def size(self) -> int:
        """
        返回视图的大小<br/>
        * 详见ScreepsArena文档

        Returns:
            int
        """
        pass

    def header(self, creep: st.creep) -> object:
        """
        为一个Creep绘制标题

        Args:
            creep: st.creep  一个Creep对象

        Returns:
            View
        """
        pass

    def cluster(self, cluster:ClusterResult, fill=None, options:UsrObject=None) -> object:
        """
        绘制聚类结果

        Args:
            cluster: ClusterResult  聚类结果
            fill: str  填充颜色 比如 '#FF0000' (优先级高于options.fill)
            options: 详见ScreepsArena文档

        Returns:
            View
        """
        pass