// > sort 6
import {LANGUAGE} from "../build/main.mjs";

class _Engine_Scheduler {
    static instance = null;
    constructor() {
        if (_Engine_Scheduler.instance !== null) {
            return _Engine_Scheduler.instance;
        }
        this.prototypes = {};  // name:prototype
        this.updates = [];
        // instance
        _Engine_Scheduler.instance = this;
    }

    login(prototype) {
        if (prototype.NAME !== undefined) {
            if (this.prototypes[prototype.NAME] !== undefined) {
                if (LANGUAGE === "cn") {
                    std.warn("Scheduler.login", "名称重复:'" + prototype.NAME + "'");
                } else {
                    std.warn("Scheduler.login", "Name repeat:'" + prototype.NAME + "'");
                }
            }
            // prototype name不允许以数字结尾
            if (prototype.NAME.match(/\d+$/)) {
                if (LANGUAGE === "cn") {
                    std.error("Scheduler.login", "类型名称不允许以数字结尾:'" + prototype.NAME + "'");
                } else {
                    std.error("Scheduler.login", "Type-Name not allowed to end with a number:'" + prototype.NAME + "'");
                }
            }
        } else if (LANGUAGE === "cn") {
            std.error("Scheduler.login", "未指定名称:'NAME'");
        } else {
            std.error("Scheduler.login", "No 'NAME' specified");
        }

        this.prototypes[prototype.NAME] = prototype;

    }

    remove(name) {
        if (_Engine_Scheduler.instance === null) return;
        var _length = len(_Engine_Scheduler.instance.updates);
        for (var i = 0; i < _length; i++) {
            var item = _Engine_Scheduler.instance.updates[i];
            if (item.py_name === name) {
                item.remove();
                _Engine_Scheduler.instance.updates.splice(i, 1);
                return;
            }
        }
    }

    add(name, ...args) {
        let instance = name;
        if (name._priority === undefined) {
            // 根据name找到prototype，然后创建实例
            const prototype = this.prototypes[name];
            if (prototype === undefined) {
                if (LANGUAGE === "cn") {
                    std.error("Scheduler.add", "找不到名称:'" + name + "'");
                } else {
                    std.error("Scheduler.add", "Name not found:'" + name + "'");
                }
                return;
            }
            try {
                instance = prototype(...args);  //  this is a python prototype, so it can be called instead of new
            } catch (e) {
                if (LANGUAGE === "cn") std.warn("Scheduler.add", "创建实例失败:'" + name + "'\n", e);
                else std.warn("Scheduler.add", "Failed to create instance:'" + name + "'\n", e);
            }
        }

        this.updates.push(instance);

        // 按权重排序
        this.updates.sort((a, b) => (b._priority || 0) - (a._priority || 0));

        return instance;
    }

    find(name) {
        /// 根据名称查找实例。如果name不以数字结尾，则返回所有该类型的实例；如果name以数字结尾，则返回对应实例。
        /// .例如，'atk'返回所有'atk'类型的实例[atk0, atk1, atk2, ...]；'atk0'仅返回'atk0'实例。
        const _length = len(this.updates);
        if (name.match(/\d+$/)) {
            for (let i = 0; i < _length; i++) {
                const item = this.updates[i];
                if (item.py_name === name) {
                    return item;
                }
            }
        } else {
            const result = [];
            for (let i = 0; i < _length; i++) {
                const item = this.updates[i];
                if (item.py_name.match(new RegExp("^" + name + "\\d+$"))) {
                    result.push(item);
                }
            }
            return result;
        }

    }

    handle() {
        if (len(this.updates) > 0) {
            let removes = null;
            let skips = null;  // 如果某item报错，跳过后续
            for (const stage of [0, 1, 2, 3]) {
                const _length = len(this.updates);
                for (let i = 0; i < _length; i++) {
                    if (skips !== null && skips.includes(i)) continue;
                    var item = this.updates[i];
                    if (item === undefined || item === null){
                        if (LANGUAGE === "cn") std.warn("Scheduler", "发现未定义或null的实例，已删除");
                        else std.warn("Scheduler", "Found undefined or null instance, deleted");
                        if (removes === null) removes = [];
                        removes.push(i);
                        continue;
                    }
                    try {
                        item.handle(stage);
                    } catch (e) {
                        if (LANGUAGE === "cn") std.warn("Scheduler", "在执行'" + item.py_name + "|stage=" + stage.toString() + "'时发生错误:\n", e, `\n    Tick=${know.now}不再处理其后续逻辑。`);
                        else std.warn("Scheduler", "An error occurred while executing '" + item.py_name + "':\n", e, `\n    Tick=${know.now} no longer handles its subsequent logic.`);
                        if (skips === null) skips = [];
                        skips.push(i);
                    }
                    if (!item._alive) {
                        item.remove();
                        if (removes === null) removes = [];
                        removes.push(i);
                    }
                }
                if (removes !== null) {
                    for (let j = len(removes) - 1; j >= 0; j--) {
                        this.updates.splice(removes[j], 1);
                    }
                    removes = null;
                }
            }
        }
    }
}

function Scheduler() {
    if (_Engine_Scheduler.instance !== null) return _Engine_Scheduler.instance;
    return new _Engine_Scheduler();
}

class _Engine_ChebRotateRef {
    // 接收一个Creep作为leader，然后允许添加一组带有相对位置信息的Creep作为follower，这些follower会跟随leader的移动而旋转
    // 默认TOP为正方向
    constructor(leader) {
        this.leader = leader;
        this.followers = [];
        this.follwer_relpos = { // name:Point

        }
        this.follower_looses = { // name:number

        }
        this._built = false;  // 如果运行，那么被设置为true

        // ---
        this.alldirs_relpos = {

        };
        this.curdir = TOP;
        this.curels = undefined;
        this._handled = false;  // 如果运行，那么被设置为true

        // getter
        this._update_tick = -1;
        this._update_pos = null;
        this.path = null;
        this.path_goal = null;
        this.path_index = 0;
        this.to = null;

    }

    get hits() {
        let hits = 0;
        if (st.creep(this.leader)) hits += this.leader.hits;
        for (const follower of this.followers) {
            if (st.creep(follower)) hits += follower.hits;
        }
        return hits;
    }

    get hitsMax() {
        let hitsMax = 0;
        if (st.creep(this.leader)) hitsMax += this.leader.hitsMax;
        for (const follower of this.followers) {
            if (st.creep(follower)) hitsMax += follower.hitsMax;
        }
        return hitsMax;
    }

    get exists() {
        const leader = this.__select_leader();
        return leader !== null && leader.exists;
    }


    get x() {
        if (this._update_tick !== know.now){
            this._update_tick = know.now;
            const leader = this.__select_leader();
            const relpos = this.__get_direlpos(leader);
            this._update_pos = Point(leader.x - relpos.x, leader.y - relpos.y);
        }
        return this._update_pos.x;
    }

    get y() {
        if (this._update_tick !== know.now){
            this._update_tick = know.now;
            const leader = this.__select_leader();
            const relpos = this.__get_direlpos(leader);
            this._update_pos = Point(leader.x - relpos.x, leader.y - relpos.y);
        }
        return this._update_pos.y;
    }

    __check_built() {
        if (this._built) {
            if (LANGUAGE === "cn") std.error("ChebRotateRef", "不允许动态添加follower");
            else std.error("ChebRotateRef", "Dynamic addition of follower is not allowed");
        }
    }

    __check_not_built() {
        if (!this._built) {
            if (LANGUAGE === "cn") std.error("ChebRotateRef", "未初始化。请确保至少handle执行一次");
            else std.error("ChebRotateRef", "Not initialized. Make sure handle is executed at least once");
        }
    }

    add(follower, relpos, loose) {
        this.__check_built();
        if (!relpos || relpos.x === undefined || relpos.y === undefined) {
            if (LANGUAGE === "cn") std.error("ChebRotateRef.add", "relpos参数不合法, relpos:", relpos.toString());
            else std.error("ChebRotateRef.add", "Illegal relpos parameter, relpos:", relpos.toString());
        }
        this.followers.push(follower);
        this.follwer_relpos[follower.id] = relpos;
        this.follower_looses[follower.id] = loose;
    }

    remove(follower) {
        this.__check_built();
        const index = this.followers.indexOf(follower);
        if (index !== -1) {
            this.followers.splice(index, 1);
            delete this.follwer_relpos[follower.id];
            delete this.follower_looses[follower.id];
        }
    }

    contains(follower) {
        return this.followers.includes(follower);
    }


    build() {
        this.__check_built();
        this._built = true;

        // 使用get.chebRotate45x(point, old_direction, new_direction) 方法，计算除Top外的其他方向的相对位置，记录到this.alldirs_relpos[TOP\TOP_RIGHT ...]中
        const directions = [TOP, TOP_RIGHT, RIGHT, BOTTOM_RIGHT, BOTTOM, BOTTOM_LEFT, LEFT, TOP_LEFT];
        this.alldirs_relpos[TOP] = this.follwer_relpos;  // 默认方向
        for (const direction of directions) {
            if (direction === TOP) continue;
            for (const follower of this.followers) {
                const relpos = this.follwer_relpos[follower.id];
                if (this.alldirs_relpos[direction] === undefined) this.alldirs_relpos[direction] = {};
                this.alldirs_relpos[direction][follower.id] = get.chebRotate45x(relpos, TOP, direction);
            }
        }
    }

    __select_leader() {
        // 选择一个leader，如果当前leader阵亡，那么选择一个follower作为leader
        if (this.leader.exists) return this.leader;
        for (const follower of this.followers) {
            if (follower.exists) return follower;
        }
        return null;
    }

    __get_direlpos(leader_follower) {
        if (this.leader.id === leader_follower.id) return Point(0, 0);
        return this.curels[leader_follower.id];   // 请检查是否已经build，视作内部函数，就不检查了
    }

    __induce_nextpos(creep) {
        // 这个函数推测目标下一步的位置
        if (creep.to !== undefined && get.moved(creep)) {
            return creep.to;
        }
        return creep;
    }

    toward(dir) {
        this.__check_not_built();
        this.curdir = dir;
        this.curels = this.alldirs_relpos[dir];
    }

    move(to){
        // 和直接put.move()的区别是，这个函数可以避免leader阵亡导致的follower无法移动的问题
        const leader = this.__select_leader();
        if (leader === null){
            if (LANGUAGE === "cn") std.error("ChebRotateRef.move", "该阵型所有单位都已阵亡");
            else std.error("ChebRotateRef.move", "All units in this formation have been wiped out");
        }
        const rel = this.__get_direlpos(leader);
        const to_pos = get.offset(to, rel);
        return put.move(leader, to_pos);
        // if (!(this.path_goal) || this.path_goal.x !== to.x && this.path_goal.y !== to.y) {
        //     var path_res = searchPath (leader, to);
        //     if (path_res.incomplete) {
        //         if (LANGUAGE === "cn") std.warn ('cheb.move', `到目标的路径不完整。`);
        //         else std.warn ('cheb.move', 'Path to target:', to, ' is incomplete.');
        //     }
        //     this.path = path_res.path;
        //     this.path_goal = to;
        //     this.path_index = 0;
        // }
        // if (this.path) {
        //     this.to = this.path [this.path_index];
        //     this.path_index++;
        //     if (this.path_index >= this.path.length) {
        //         this.path = null;
        //         this.path_index = 0;
        //     }
        // }
    }

    handle(){
        //这个函数根据当前leader的位置和this.curdir，更新所有follower的位置并返回
        if (!this._built) this.build();

        // leader.x, leader.y;
        // 如果leader阵亡，那么寻找一个依然存在的follower，计算出虚拟的leader位置
        const leader = this.__select_leader();
        if (leader === null) {
            if (LANGUAGE === "cn") std.error("ChebRotateRef.handle", "该阵型所有单位都已阵亡");
            else std.error("ChebRotateRef.handle", "All units in this formation have been wiped out");
        }
        const leader_rel = this.__get_direlpos(leader);
        // const leader_pos = this.__induce_nextpos(leader);
        const leader_pos = leader;

        // 获得next的相对于原始leader的位置
        const virtual_leader = {
            x: leader_pos.x - leader_rel.x,
            y: leader_pos.y - leader_rel.y,
        }

        // 先遍历一轮，检查所有follower的loose情况
        if (this._handled)
            for (const follower of this.followers) {
                if (get.terrain(follower.refpos) >= TERRAIN_WALL) continue;
                if (leader.x ===  follower.x && leader.y === follower.y) continue;
                const blocker = get.find(this.followers, c => c !== follower && c.x === follower.refpos.x && c.y === follower.refpos.y);
                if (blocker) continue;
                const dist = follower.distance(follower.refpos);
                if (dist > this.follower_looses[follower.id]) {
                    // 暂停更新
                    return;
                }
            }
        else
            this._handled = true;


        leader.refpos = virtual_leader;
        // console.log(this.curels);
        for (const follower of this.followers) {
            const relpos = this.curels[follower.id];
            follower.refpos = Point(virtual_leader.x + relpos.x, virtual_leader.y + relpos.y);
            // print(`follower ${follower} set ref: ${follower.refpos.x}, ${follower.refpos.y}`)
        }
    }
}

function ChebRotateRef(leader) {
    return new _Engine_ChebRotateRef(leader);
}

class _Engine_Monitor {
    static instance = null;
    constructor(N) {
        if (_Engine_Monitor.instance !== null) {
            return _Engine_Monitor.instance;
        }
        if (N < 2) {
            if  (LANGUAGE === "cn") std.error("Monitor", "参数 N 必须大于等于 2:", N);
            else std.error("Monitor", "Parameter N must be greater than or equal to 2:", N);
        }
        this.N = N;
        this.olds = null;
        this.nows = null;
        this.name= "monitor";
        this.py_name= "monitor";
        this.eda = new ListenSystem(this);
        // --------
        this.nowBatchSize = 0;
        this.oldBatchSize = 0;
        this.nowCurrentIndex = 0; // 当前正在处理的gobject的索引
        this.oldCurrentIndex = 0;
        // --------
        this.creepHistory = {
            enemies:{
                vector: new PartsVector([]),
                count: 0,
                grade: 0,
            },
            friends:{
                vector: new PartsVector([]),
                count: 0,
                grade: 0,
            }
        }
        // -----------
        this.records = {

        }
    }

    onCreepAdded(creeps){
        const enemies = [];
        const friends = [];
        let enemies_grade = 0;
        let friends_grade = 0;
        const enemies_partsVector = new PartsVector([]);
        const friends_partsVector = new PartsVector([]);
        for (const creep of creeps) {
            if (!creep.my){
                enemies.push(creep);
                enemies_grade += creep.grade;
                enemies_partsVector.add(creep.partsVector);
            } else{
                friends.push(creep);
                friends_grade += creep.grade;
                friends_partsVector.add(creep.partsVector);
            }
        }
        if (enemies.length > 0) {
            this.creepHistory.enemies.vector.add(enemies_partsVector);
            this.creepHistory.enemies.count += enemies.length;
            this.creepHistory.enemies.grade += enemies_grade;
            this.eda.pushEvent("enemyNew:Creep", {
                creeps: enemies,
                current: {
                    grade: enemies_grade,
                    count: enemies.length,
                    vector: enemies_partsVector
                },
                history: this.creepHistory
            });
        }
        if (friends.length > 0) {
            this.creepHistory.friends.vector.add(friends_partsVector);
            this.creepHistory.friends.count += friends.length;
            this.creepHistory.friends.grade += friends_grade;
            this.eda.pushEvent("myNew:Creep", {
                creeps: friends,
                current: {
                    grade: enemies_grade,
                    count: friends.length,
                    vector: friends_partsVector
                },
                history: this.creepHistory
            });
        }
        
    }

    onCreepRemoved(creeps){
        const enemies = [];
        const friends = [];
        let enemies_grade = 0;
        let friends_grade = 0;
        const enemies_partsVector = new PartsVector([]);
        const friends_partsVector = new PartsVector([]);
        for (const creep of creeps){
            if (!creep.my){
                enemies.push(creep);
                enemies_grade += creep.grade;
                enemies_partsVector.add(creep.partsVector);
            } else{
                friends.push(creep);
                friends_grade += creep.grade;
                friends_partsVector.add(creep.partsVector);
            }

            delete creep.py;
        }
        // print("enemyLost:", enemies.length, enemies_grade)
        // print("myLost:", friends.length, friends_grade)
        if (enemies.length > 0) {
            this.creepHistory.enemies.vector.sub(enemies_partsVector);
            this.creepHistory.enemies.count -= enemies.length;
            this.creepHistory.enemies.grade -= enemies_grade;
            this.eda.pushEvent("enemyLost:Creep", {
                creeps: enemies,
                current: {
                    count: enemies.length,
                    grade: enemies_grade,
                    vector: enemies_partsVector,
                },
                history: this.creepHistory
            });
        }
        if (friends.length > 0) {
            this.creepHistory.friends.vector.sub(friends_partsVector);
            this.creepHistory.friends.count -= friends.length;
            this.creepHistory.friends.grade -= friends_grade;
            this.eda.pushEvent("myLost:Creep", {
                creeps: friends,
                current: {
                    count: friends.length,
                    grade: friends_grade,
                    vector: friends_partsVector,
                },
                history: this.creepHistory
            });
        }
    }

    onWallAdded(walls){
        if (this.records["Wall"] === undefined) {
            this.records["Wall"] = [];
        }
        for (const wall of walls) {
            COST_MATRIX_LIBS.default.set(wall.x, wall.y, 255);
            COST_MATRIX_LIBS.swamp.set(wall.x, wall.y, 255);
            get.COST_ARRAY2D[wall.y][wall.x] = 255;
            this.records["Wall"].push({id: wall.id, x: wall.x, y: wall.y});
        }


    }

    onExtensionAdded(exts){
        if (this.records["Extension"] === undefined) {
            this.records["Extension"] = [];
        }
        for (const ext of exts) {
            COST_MATRIX_LIBS.default.set(ext.x, ext.y, 255);
            COST_MATRIX_LIBS.swamp.set(ext.x, ext.y, 255);
            get.COST_ARRAY2D[ext.y][ext.x] = 255;
            this.records["Extension"].push({id: ext.id, x: ext.x, y: ext.y});
        }


    }

    onSpawnAdded(spawns){
        if (this.records["Spawn"] === undefined) {
            this.records["Spawn"] = [];
        }
        for (const spawn of spawns) {
            COST_MATRIX_LIBS.default.set(spawn.x, spawn.y, 255);
            COST_MATRIX_LIBS.swamp.set(spawn.x, spawn.y, 255);
            get.COST_ARRAY2D[spawn.y][spawn.x] = 255;
            this.records["Spawn"].push({id: spawn.id, x: spawn.x, y: spawn.y});
        }

    }

    onTowerAdded(towers){
        if (this.records["Tower"] === undefined) {
            this.records["Tower"] = [];
        }
        for (const tower of towers) {
            COST_MATRIX_LIBS.default.set(tower.x, tower.y, 255);
            COST_MATRIX_LIBS.swamp.set(tower.x, tower.y, 255);
            get.COST_ARRAY2D[tower.y][tower.x] = 255;
            this.records["Tower"].push({id: tower.id, x: tower.x, y: tower.y});
        }

    }
    
    
    onRoadAdded(roads){
        if (this.records["Road"] === undefined) {
            this.records["Road"] = [];
        }
        for (const road of roads) {
            COST_MATRIX_LIBS.default.set(road.x, road.y, 1);
            COST_MATRIX_LIBS.swamp.set(road.x, road.y, 1);
            get.COST_ARRAY2D[road.y][road.x] = 1;
            this.records["Road"].push({id: road.id, x: road.x, y: road.y});
        }

    }


    onRampartAdded(ramps){
        if (this.records["Rampart"] === undefined) {
            this.records["Rampart"] = [];
        }
        for (const ramp of ramps) {
            if (ramp.my) continue;
            COST_MATRIX_LIBS.default.set(ramp.x, ramp.y, 255);
            COST_MATRIX_LIBS.swamp.set(ramp.x, ramp.y, 255);
            get.COST_ARRAY2D[ramp.y][ramp.x] = 255;
            this.records["Rampart"].push({id: ramp.id, x: ramp.x, y: ramp.y});
        }

    }

    onWallRemoved(walls){
        if (this.records["Wall"] === undefined || this.records["Wall"].length === 0) return;
        for (let wall of walls) {
            // 从records中弹出该元素
            const index = this.records["Wall"].findIndex((record) => record.id === wall.id);
            if (index !== -1) {
                wall = this.records["Wall"][index];
                this.records["Wall"].splice(index, 1);
                const terrain = get.terrain(wall.x, wall.y);
                COST_MATRIX_LIBS.default.set(wall.x, wall.y, terrain);
                COST_MATRIX_LIBS.swamp.set(wall.x, wall.y, terrain);
                get.COST_ARRAY2D[wall.y][wall.x] = terrain;
            }
        }
    }

    onExtensionRemoved(exts){
        if (this.records["Extension"] === undefined || this.records["Extension"].length === 0) return;
        for (let ext of exts) {
            // 从records中弹出该元素
            const index = this.records["Extension"].findIndex((record) => record.id === ext.id);
            if (index !== -1) {
                ext = this.records["Extension"][index];
                this.records["Extension"].splice(index, 1);
                const terrain = get.terrain(ext.x, ext.y);
                COST_MATRIX_LIBS.default.set(ext.x, ext.y, terrain);
                COST_MATRIX_LIBS.swamp.set(ext.x, ext.y, terrain);
                get.COST_ARRAY2D[ext.y][ext.x] = terrain;
            }
        }
    }

    onSpawnRemoved(spawns){
        if (this.records["Spawn"] === undefined || this.records["Spawn"].length === 0) return;
        for (let spawn of spawns) {
            // 从records中弹出该元素
            const index = this.records["Spawn"].findIndex((record) => record.id === spawn.id);
            if (index !== -1) {
                spawn = this.records["Spawn"][index];
                this.records["Spawn"].splice(index, 1);
                const terrain = get.terrain(spawn.x, spawn.y);
                COST_MATRIX_LIBS.default.set(spawn.x, spawn.y, terrain);
                COST_MATRIX_LIBS.swamp.set(spawn.x, spawn.y, terrain);
                get.COST_ARRAY2D[spawn.y][spawn.x] = terrain;
            }
        }
    }

    onTowerRemoved(towers){
        if (this.records["Tower"] === undefined || this.records["Tower"].length === 0) return;
        for (let tower of towers) {
            // 从records中弹出该元素
            const index = this.records["Tower"].findIndex((record) => record.id === tower.id);
            if (index !== -1) {
                tower = this.records["Tower"][index];
                this.records["Tower"].splice(index, 1);
                const terrain = get.terrain(tower.x, tower.y);
                COST_MATRIX_LIBS.default.set(tower.x, tower.y, terrain);
                COST_MATRIX_LIBS.swamp.set(tower.x, tower.y, terrain);
                get.COST_ARRAY2D[tower.y][tower.x] = terrain;
            }
        }
    }

    onRoadRemoved(roads){
        if (this.records["Road"] === undefined || this.records["Road"].length === 0) return;
        for (let road of roads) {
            // 从records中弹出该元素
            const index = this.records["Road"].findIndex((record) => record.id === road.id);
            if (index !== -1) {
                road = this.records["Road"][index];
                this.records["Road"].splice(index, 1);
                const terrain = get.terrain(road.x, road.y);
                COST_MATRIX_LIBS.default.set(road.x, road.y, terrain);
                COST_MATRIX_LIBS.swamp.set(road.x, road.y, terrain);
                get.COST_ARRAY2D[road.y][road.x] = terrain;
            }
        }
    }

    onRampartRemoved(ramps){
        if (this.records["Rampart"] === undefined || this.records["Rampart"].length === 0) return;
        for (let ramp of ramps) {
            // 从records中弹出该元素
            const index = this.records["Rampart"].findIndex((record) => record.id === ramp.id);
            if (index !== -1) {
                ramp = this.records["Rampart"][index];
                this.records["Rampart"].splice(index, 1);
                const terrain = get.terrain(ramp.x, ramp.y);
                COST_MATRIX_LIBS.default.set(ramp.x, ramp.y, terrain);
                COST_MATRIX_LIBS.swamp.set(ramp.x, ramp.y, terrain);
                get.COST_ARRAY2D[ramp.y][ramp.x] = terrain;
            }
        }
    }

    onNewCallback(news){
        // news是一个{type-string: array<gobject>}
        for (const type in news) {
            this.eda.pushEvent(`__new:${type}__`, {
                objects: news[type],
            })
            // for (const obj of news[type]) {
            //     print("new:", obj, type)
            // }
            if (type === "Creep") this.onCreepAdded(news[type]);
            else if (type === "StructureWall") this.onWallAdded(news[type]);
            else if (type === "StructureSpawn") this.onSpawnAdded(news[type]);
            else if (type === "StructureTower") this.onTowerAdded(news[type]);
            else if (type === "StructureRoad") this.onRoadAdded(news[type]);
            else if (type === "StructureRampart") this.onRampartAdded(news[type]);
            else if (type === "StructureExtension") this.onExtensionAdded(news[type]);
        }
    }
    onLostCallback(losts){
        // losts是一个{type-string: array<gobject>}
        for (const type in losts) {
            this.eda.pushEvent(`__lost:${type}__`, {
                objects: losts[type],
            })
            if (type === "Creep") this.onCreepRemoved(losts[type]);
            else if (type === "StructureWall") this.onWallRemoved(losts[type]);
            else if (type === "StructureSpawn") this.onSpawnRemoved(losts[type]);
            else if (type === "StructureTower") this.onTowerRemoved(losts[type]);
            else if (type === "StructureRoad") this.onRoadRemoved(losts[type]);
            else if (type === "StructureRampart") this.onRampartRemoved(losts[type]);
            else if (type === "StructureExtension") this.onExtensionRemoved(losts[type]);
        }
    }

    calcBatchSize() {
        // 假设新版的逻辑是，第一个tick处理平常tick 50%的工作量，后续tick处理量为100%.  所有tick的总工作量恰好等于一轮完整的工作量
        const total = (this.N - 1) * 2 + 1;
        this.oldBatchSize = Math.ceil(this.olds.length / total * 2);
        this.nowBatchSize = Math.ceil(this.nows.length / total * 2);
    }

    partialHandle(batch_count){
        const news = {};
        const losts = {};
        let any_news_flag = false;
        let any_losts_flag = false;

        /// 遵循左闭右开原则

        // Now
        const nowEndIndex = Math.min(
            this.nowCurrentIndex + Math.ceil(batch_count * this.nowBatchSize),
            this.nows.length
        );
        for (let i = this.nowCurrentIndex; i < nowEndIndex; i++) {
            const now = this.nows[i];
            // check new
            let flag = false;
            for (const old of this.olds) {
                if (old.id === now.id) {
                    flag = true;
                    break;
                }
            }
            if (!flag) {
                const typename = now.constructor.name;
                if (news[typename] === undefined) {
                    news[typename] = [];
                }
                news[typename].push(now);
                any_news_flag = true;
            }
        }

        // Old
        const oldEndIndex = Math.min(
            this.oldCurrentIndex + Math.ceil(batch_count * this.oldBatchSize),
            this.olds.length
        );
        for (let i = this.oldCurrentIndex; i < oldEndIndex; i++) {
            const old = this.olds[i];
            // check lost
            let flag = false;
            for (const now of this.nows) {
                if (old.id === now.id) {
                    flag = true;
                    break;
                }
            }
            if (!flag) {
                const typename = old.constructor.name;
                if (losts[typename] === undefined) {
                    losts[typename] = [];
                }
                losts[typename].push(old);
                any_losts_flag = true;
            }
        }

        if (any_news_flag) this.onNewCallback(news); // 处理new
        if (any_losts_flag) this.onLostCallback(losts); // 处理lost

        // 更新索引
        this.nowCurrentIndex = nowEndIndex;
        this.oldCurrentIndex = oldEndIndex;
    }
    handle() {
        const state = (know.now - 1) % this.N;  // run at tick==1
        // iter all creep
        DEFAULT_MOTION.costMatrix = COST_MATRIX_LIBS.default.clone();
        SWAMP_MOTION.costMatrix = COST_MATRIX_LIBS.swamp.clone();
        for (const creep of get.creeps()) {
            DEFAULT_MOTION.costMatrix.set(creep.x, creep.y, 254);
            SWAMP_MOTION.costMatrix.set(creep.x, creep.y, 254);
        }
        if (state === 0 || this.olds === null) {
            this.olds = this.nows;
            this.nows = get.gobjects();
            if (this.olds === null) {
                this.olds = this.nows;
                const walls = get.filter(this.olds, c=>c instanceof StructureWall);
                this.onWallAdded(walls);
                const extensions = get.filter(this.olds, c=>c instanceof StructureExtension);
                this.onExtensionAdded(extensions);
                const spawns = get.filter(this.olds, c=>c instanceof StructureSpawn);
                this.onSpawnAdded(spawns);
                const towers = get.filter(this.olds, c=>c instanceof StructureTower);
                this.onTowerAdded(towers);
            }
            this.calcBatchSize();
            this.nowCurrentIndex = 0;
            this.oldCurrentIndex = 0;

            this.partialHandle(0.5);
        } else {
            this.partialHandle(1);
        }

    }
}

function Monitor(N) {
    if (_Engine_Monitor.instance === null) {
        _Engine_Monitor.instance = new _Engine_Monitor(N);
    }
    return _Engine_Monitor.instance;
}
