// > sort 6

class __View {
    static DEFAULT_OPTS = {
        font: '0.5',
        align: 'center',
        opacity: 0.7,
        color: '#AEFC80'
    }
    constructor(layer, persistent = true) {
        this.v = new Visual(layer, persistent);
    }

    clear() {
        this.v.clear(); // > ignore
        return this;
    }

    text(text, pos, font, color, options = null) {
        options = options? options : {};
        options.font = font;
        options.color = color;
        this.v.text(text, pos, options);
        return this;
    }

    circle(pos, radius, fill = null, options = null) {
        options = options? options : {};
        options.radius = radius;
        options.fill = fill;
        this.v.circle(pos, options);
        return this;
    }

    line(pos1, pos2, width = 0.1, color = null, options = null) {
        options = options? options : {};
        options.width = width;
        options.color = color;
        this.v.line(pos1, pos2, options);
        return this;
    }

    rect(pos, width, height, fill = null, options = null) {
        options = options? options : {};
        options.fill = fill;
        this.v.rect(pos, width, height, options);
        return this;
    }

    poly(points, fill = null, options = null) {
        options = options? options : {};
        options.fill = fill;

        let array = [];
        for (let i = 0; i < points.length; i++) {
            array.push(points[i]);
        }

        this.v.poly(array, options);
        return this;
    }

    size() {
        return this.v.size();
    }

    header(creep) {
        if (creep) {
            let txt = creep.name || creep.id;
            let hp_percent = creep.hits / creep.hitsMax * 100;  // int:0 - 100
            let _hex = Math.floor(hp_percent * 255 / 100).toString(16).padStart(2, '0');
            let color = `#AE${_hex}80`;
            txt += `❤️${Math.round(hp_percent)}`;
            this.text(txt, creep.offset(0, -0.6), 0.4, color, __View.DEFAULT_OPTS);
        } else {
            std.warn("View.header", "No creep object:", creep);
        }
        return this;
    }

    cluster(cluster, fill = null, options = null) {
        options = options ?? {}; // 使用空值合并运算符更简洁
        options.fill = fill || __View.DEFAULT_OPTS.color;
        options.opacity = 0.2;

        const grid = new Map(); // 使用 Map 存储坐标及其相关内容

        // 遍历 cluster.objects，计算 7x7 格子并存入 Map
        for (const obj of cluster.objects) {
            for (let i = -3; i <= 3; i++) {
                for (let j = -3; j <= 3; j++) {
                    const key = `${obj.x + i},${obj.y + j}`; // 用字符串 "x,y" 作为键
                    grid.set(key, true); // 存储任何值（这里用 true 表示存在）
                }
            }
        }

        // 遍历 Map，将每个格子画成方形
        for (const key of grid.keys()) {
            const [x, y] = key.split(',').map(Number);
            this.rect(
                { x: x - 0.5, y: y - 0.5 },
                1,
                1,
                fill,
                options
            );
        }

        // 在cluster的中心位置绘制cluster的grade:int信息
        this.text(`grade:${cluster.grade}`, Point(cluster.x, cluster.y), 0.4, __View.DEFAULT_OPTS.color, __View.DEFAULT_OPTS);
    }
}

function View(layer, persistent = true) {
    return new __View(layer, persistent);
}
